"""Top level interfaces and definitions for Lakshmi."""

from lakshmi.assets import FromDict, ToDict
from lakshmi.table import Table
import lakshmi.utils as utils
import yaml


class Account:
    """Class representing an account."""
    def __init__(self, name, account_type):
        """
        Arguments:
          name: Printable name for this account.
          account_type: Type of this account.
        """
        self._name = name
        self.account_type = account_type
        self._assets = {}
        self._cash = 0

    def ToDict(self):
        d = {'Name': self._name,
             'Account Type': self.account_type}
        if self._assets:
             d['Assets'] = [ToDict(asset) for asset in self._assets.values()]
        if self._cash != 0:
            d['Available Cash'] = self._cash
        return d

    @classmethod
    def FromDict(cls, d):
        ret_obj = Account(d.pop('Name'), d.pop('Account Type'))
        for asset_dict in d.pop('Assets', []):
            ret_obj.AddAsset(FromDict(asset_dict))
        ret_obj._cash = d.pop('Available Cash', 0)
        assert len(d) == 0, 'Extra attributes found: ' + str(list(d.keys()))
        return ret_obj

    def String(self):
        table = Table(2)
        table.AddRow(['Name:', f'{self._name}'])
        table.AddRow(['Type:', f'{self.account_type}'])
        total = sum([asset.AdjustedValue() for asset in self._assets.values()])
        table.AddRow(['Total:', utils.FormatMoney(total)])
        if self._cash:
            table.AddRow(['Available Cash:', utils.FormatMoneyDelta(self._cash)])
        return table.String(tablefmt='plain')

    def AddAsset(self, asset, replace=False):
        assert replace or asset.ShortName() not in self._assets, (
            f'Attempting to add duplicate Asset: {asset.ShortName()}')
        self._assets[asset.ShortName()] = asset
        return self

    def Assets(self):
        return self._assets.values()

    def SetAssets(self, assets):
        self._assets = {}
        for asset in assets:
            self.AddAsset(asset)

    def GetAsset(self, short_name):
        return self._assets[short_name]

    def RemoveAsset(self, short_name):
        del self._assets[short_name]

    def Name(self):
        return self._name

    def AddCash(self, delta):
        self._cash += delta

    def AvailableCash(self):
        return self._cash


class AssetClass:
    """(Tree of) Asset classes."""
    def __init__(self, name):
        self.name = name
        self.children = []
        # Populated when _Validate is called.
        self._leaves = None

    def ToDict(self):
        d = {'Name': self.name}
        if self.children:
            d['Children'] = [{'Ratio': ratio} | child.ToDict()
                             for child, ratio in self.children]
        return d

    @classmethod
    def FromDict(cls, d):
        ret_obj = AssetClass(d.pop('Name'))
        for child_dict in d.pop('Children', []):
            ret_obj.AddSubClass(
                child_dict.pop('Ratio'),
                AssetClass.FromDict(child_dict))
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj.Validate()

    def Copy(self):
        """Returns a copy of this AssetClass and its sub-classes."""
        ret_val = AssetClass(self.name)
        for child, ratio in self.children:
            ret_val.AddSubClass(ratio, child.Copy())
        return ret_val

    def AddSubClass(self, ratio, asset_class):
        self.children.append((asset_class, ratio))
        # Leaves is not upto date now, need validation again.
        self._leaves = None
        return self

    def _Validate(self):
        """Returns a tuple (leaf names, class names) for the subtree."""
        # Check if all percentages add up to 100%
        if not self.children:
            self._leaves = {self.name}
            return self._leaves, [self.name]

        self._leaves = set()
        class_names = [self.name]
        total = 0.0
        for asset_class, ratio in self.children:
            assert ratio >= 0.0 and ratio <= 1.0, (
                    f'Bad ratio provided to Asset Class ({ratio})')
            total += ratio
            temp_leafs, temp_classes = asset_class._Validate()
            self._leaves.update(temp_leafs)
            class_names += temp_classes

        assert abs(total - 1) < 1e-6, (
            f'Sum of sub-classes is not 100% (actual: {total * 100}%)')

        return self._leaves, class_names

    def Validate(self):
        unused_leaves, all_class_names = self._Validate()
        duplicates = set([x for x in all_class_names
                          if all_class_names.count(x) > 1])

        assert not duplicates, f'Found duplicate Asset class(es): {duplicates}'
        return self

    def _Check(self):
        assert self._leaves, 'Need to validate AssetAllocation before using it.'

    def FindAssetClass(self, asset_class_name):
        """Returns a tuple of object representing asset_class_name and its desired ratio.

        Returns None if asset_class_name is not found."""
        self._Check()

        if self.name == asset_class_name:
            return self, 1.0

        for asset_class, ratio in self.children:
            ret_value = asset_class.FindAssetClass(asset_class_name)
            if ret_value:
                return ret_value[0], ret_value[1] * ratio

        return None

    def Leaves(self):
        self._Check()
        return self._leaves

    def ValueMapped(self, money_allocation):
        """Returns how much money is mapped to this Asset Class or it's children.

        Arguments:
          money_allocation: A map of leaf_class_name -> money.
        """
        self._Check()
        return sum([value for name, value in money_allocation.items()
                    if name in self._leaves])

    class Allocation:
        class Children:
            def __init__(self, name, actual_allocation, desired_allocation,
                    value, value_difference):
                self.name = name
                self.actual_allocation = actual_allocation
                self.desired_allocation = desired_allocation
                self.value = value
                self.value_difference = value_difference

        def __init__(self, name, value):
            self.name = name
            self.value = value
            self.children = []

        def AddChild(self, name, actual, desired):
            self.children.append(
                self.Children(name, actual, desired, actual * self.value,
                    (desired - actual) * self.value))

    def ReturnAllocation(self, money_allocation, levels=-1):
        """Returns actual and desired allocation based on how money is allocated.

        Arguments:
          money_allocation: A map of leaf_class_name -> money.
          levels: How many levels of child allocation to return (-1 = all).

        Returns:
        A list of ActualAllocation objects (for itself and any child classes
        based on the levels flag).
        """
        value = self.ValueMapped(money_allocation)
        actual_alloc = self.Allocation(self.name, value)

        if value == 0:
            return [actual_alloc]

        for asset_class, desired_ratio in self.children:
            actual_ratio = asset_class.ValueMapped(money_allocation) / value
            actual_alloc.AddChild(
                asset_class.name,
                actual_ratio,
                desired_ratio)

        ret_val = [actual_alloc]

        if levels == 0:
            return ret_val

        if levels > 0:
            levels -= 1

        for asset_class, unused_ratio in self.children:
            ret_val += asset_class.ReturnAllocation(money_allocation, levels)

        return ret_val


class Portfolio:
    def __init__(self, asset_classes):
        self._accounts = {}
        self.asset_classes = asset_classes.Validate()
        self._leaf_asset_classes = asset_classes.Leaves()

    def Save(self, filename):
        with open(filename, 'w') as f:
            yaml.dump(self.ToDict(), f, sort_keys=False)

    @classmethod
    def Load(cls, filename):
        with open(filename) as f:
            d = yaml.load(f.read(), Loader=yaml.SafeLoader)
        return Portfolio.FromDict(d)

    def ToDict(self):
        d = {'Asset Classes': self.asset_classes.ToDict()}
        if self._accounts:
            d['Accounts'] = [account.ToDict()
                             for account in self._accounts.values()]
        return d

    @classmethod
    def FromDict(cls, d):
        ret_obj = Portfolio(AssetClass.FromDict(d.pop('Asset Classes')))
        for account_dict in d.pop('Accounts', []):
            ret_obj.AddAccount(Account.FromDict(account_dict))
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def AddAccount(self, account, replace=False):
        for asset in account.Assets():
            for asset_class in asset.class2ratio.keys():
                assert asset_class in self._leaf_asset_classes, (
                    f'Unknown or non-leaf asset class: {asset_class}')

        assert replace or account.Name() not in self._accounts, (
            f'Attempting to add duplicate account: {account.Name()}')

        self._accounts[account.Name()] = account
        return self

    def RemoveAccount(self, account_name):
        del self._accounts[account_name]

    def Accounts(self):
        return self._accounts.values()

    def GetAccount(self, name):
        return self._accounts[name]

    def GetAccountNameBySubStr(self, account_str):
        """Returns account name who name partially matches account_str.

        This method throws an AssertionError if more than one or none of the
        accounts match the account_str.
        """

        matched_accounts = list(filter(lambda x: x.count(account_str),
                [account.Name() for account in self.Accounts()]))
        if len(matched_accounts) == 0:
            raise AssertionError(f'{account_str} does not match any account '
                    'in the portfolio')
        if len(matched_accounts) > 1:
            raise AssertionError(f'{account_str} matches more than one '
                        'account in the portfolio')
        return matched_accounts[0]

    def GetAssetNameBySubStr(self, account_str='', asset_str=''):
        """Returns a tuple account_name, asset_name where account name
        partially matches account_str and asset name partially matches
        asset_str or asset short name == asset_str
        Raises AssertionError if none or more than one asset matches the
        sub-strings."""
        matched_assets = list(filter(
                lambda x: x[0].count(account_str) and (
                    x[1].count(asset_str) or x[2] == asset_str),
                [(account.Name(), asset.Name(), asset.ShortName())
                    for account in self.Accounts()
                    for asset in account.Assets()]))
        if len(matched_assets) < 1:
            raise AssertionError('Provided asset and account strings match none '
                    'of the assets.')
        if len(matched_assets) > 1:
            raise AssertionError('Provided asset and account strings match more '
                    'than one assets.')
        return matched_assets[0][0], matched_assets[0][2]

    def WhatIf(self, account_name, asset_name, delta):
        """Runs a whatif scenario if asset_name in account_name is changed by delta."""
        account = self.GetAccount(account_name)
        asset = account.GetAsset(asset_name)
        asset.WhatIf(delta)
        # We take the money out of account.
        account.AddCash(-delta)

    def WhatIfAddCash(self, account_name, cash_delta):
        self.GetAccount(account_name).AddCash(cash_delta)

    def GetWhatIfs(self):
        account_whatifs = Table(
            2,
            headers=['Account', 'Cash'],
            coltypes=['str', 'delta_dollars'])
        asset_whatifs = Table(
            3,
            headers=['Account', 'Asset', 'Delta'],
            coltypes=['str', 'str', 'delta_dollars'])

        for account in self.Accounts():
            if account.AvailableCash() != 0.0:
                account_whatifs.AddRow(
                    [account.Name(), account.AvailableCash()])
            for asset in account.Assets():
                delta = asset.AdjustedValue() - asset.Value()
                if delta != 0.0:
                    asset_whatifs.AddRow([account.Name(), asset.Name(), delta])

        return account_whatifs, asset_whatifs

    def ResetWhatIfs(self):
        for account in self.Accounts():
            account.AddCash(-account.AvailableCash())
            for asset in account.Assets():
                asset.WhatIf(asset.Value() - asset.AdjustedValue())

    def TotalValue(self):
        """Returns total of all assets added."""
        total = 0.0
        for account in self.Accounts():
            total += account.AvailableCash()
            for asset in account.Assets():
                total += asset.AdjustedValue()
        return total

    def Assets(self, short_name=False, quantity=False):
        """Returns all the assets."""
        table = Table(
                3 + short_name + quantity,
                headers=['Account'] +
                        (['Name'] if short_name else []) +
                        (['Quantity'] if quantity else []) +
                        ['Asset', 'Value'],
                coltypes=['str'] +
                         (['str'] if short_name else []) +
                         (['float'] if quantity else []) +
                         ['str', 'dollars'])
        for account in self.Accounts():
            for asset in account.Assets():
                row = ([account.Name()] +
                        ([f'{asset.ShortName()}'] if short_name else []) +
                        [asset.Name(), asset.AdjustedValue()])
                if quantity:
                    row.insert(1 + short_name, asset.shares
                            if hasattr(asset, 'shares') else None)
                table.AddRow(row)

        return table

    def AssetLocation(self):
        """Returns asset location as a list of [asset class, account_type, percentage, value]."""
        class2type = {}  # Mapping of asset class -> account type -> money
        for account in self.Accounts():
            for asset in account.Assets():
                for name, ratio in asset.class2ratio.items():
                    if not name in class2type:
                        class2type[name] = {}
                    class2type[name][account.account_type] = class2type[name].get(
                            account.account_type, 0) + ratio * asset.AdjustedValue()

        table = Table(
                4,
                headers=['Asset Class', 'Account Type', 'Percentage', 'Value'],
                coltypes=['str', 'str', 'percentage', 'dollars'])

        for asset_class, type2value in class2type.items():
            first = True
            total = sum(type2value.values())
            if abs(total) < 1e-6:
                continue
            for account_type, value in sorted(type2value.items(), key=lambda x: x[1], reverse=True):
                table.AddRow([asset_class if first else '', account_type,
                    value / total, value])
                first = False
        return table

    def _GetAssetClassToValue(self):
        asset_class_to_value = {}

        for account in self.Accounts():
            for asset in account.Assets():
                for name, ratio in asset.class2ratio.items():
                    asset_class_to_value[name] = asset_class_to_value.get(
                        name, 0) + ratio * asset.AdjustedValue()

        return asset_class_to_value

    def AssetAllocationTree(self, levels=-1):
        table = Table(4,
                      headers=['Class', 'Actual%', 'Desired%', 'Value'],
                      coltypes=['str', 'percentage', 'percentage', 'dollars'])
        first_row = True
        for alloc in self.asset_classes.ReturnAllocation(
                self._GetAssetClassToValue(), levels):
            if not alloc.children:
                continue

            if not first_row:
                table.AddRow([' '])
            first_row = False
            table.AddRow([f'{alloc.name}:'])
            for child in alloc.children:
                table.AddRow([child.name,
                              child.actual_allocation,
                              child.desired_allocation,
                              child.value])
        return table

    def AssetAllocation(self, asset_class_list):
        flat_asset_class = AssetClass('Root')
        for asset_class in asset_class_list:
            found = self.asset_classes.FindAssetClass(asset_class)
            assert found, f'Could not find {asset_class}'
            flat_asset_class.AddSubClass(found[1], found[0])

        try:
            flat_asset_class.Validate()
        except AssertionError:
            raise AssertionError('AssetAllocation called with overlapping '
                    'Asset Classes or Asset Classes which does not cover the '
                    'full tree.') from None

        alloc = flat_asset_class.ReturnAllocation(
            self._GetAssetClassToValue(), 0)[0]
        table = Table(
            5,
            headers=['Class', 'Actual%', 'Desired%', 'Value', 'Difference'],
            coltypes=['str', 'percentage', 'percentage', 'dollars',
                      'delta_dollars'])
        for child in alloc.children:
            table.AddRow([child.name, child.actual_allocation,
                          child.desired_allocation, child.value,
                          child.value_difference])
        return table

    def AssetAllocationCompact(self):
        """Prints a 'compact' version of AA (Authors' favorite way of viewing his AA)."""
        def FindIndex(class_name, ret_list):
            # We assume ret_list has atleast one entry.
            names_list = [row[-3] for row in ret_list]
            return names_list.index(class_name)

        ret_list = []
        for alloc in self.asset_classes.ReturnAllocation(
                self._GetAssetClassToValue()):
            if not alloc.children:
                continue

            if not ret_list:
                for child in alloc.children:
                    ret_list.append([child.name,
                                     child.actual_allocation,
                                     child.desired_allocation])
            else:  # Parent is already in ret_list
                index = FindIndex(alloc.name, ret_list)
                # We know that here is atleast one child.
                for i in range(len(alloc.children) - 1):
                    # Make room for rest of the children by inserting empty extra rows of the
                    # same size as the parent's row
                    ret_list.insert(index + 1, [None] * len(ret_list[index]))
                for i in range(len(alloc.children)):
                    ret_list[index + i].extend(
                        [alloc.children[i].name,
                         alloc.children[i].actual_allocation,
                         alloc.children[i].desired_allocation])

        # Return early if there is no AA.
        if not ret_list:
            return Table(0)

        # By this time ret_list has all the AA tree.
        # Add Leaf node AA at the end.
        cols = max(map(len, ret_list))
        leaf_aa = self.AssetAllocation(self.asset_classes.Leaves())
        for leaf_row in leaf_aa.List():  # Format: Class, A%, D%, Value, Diff
            ret_list_row = ret_list[FindIndex(leaf_row[0], ret_list)]
            ret_list_row.extend([None] * (cols - len(ret_list_row)))
            ret_list_row.extend(leaf_row[1:])

        # All done, now build the table.
        t = Table(cols + 4,
                  headers=['Class', 'A%', 'D%'] * int(cols / 3) +
                  leaf_aa.Headers()[1:],
                  coltypes=['str', 'percentage', 'percentage'] * int(cols / 3) +
                  ['percentage', 'percentage', 'dollars', 'delta_dollars'])
        t.SetRows(ret_list)
        return t



from abc import ABC, abstractmethod
from datetime import datetime
import functools
from hashlib import md5
from pathlib import Path
import pickle

# Inspired by https://pypi.org/project/cache-to-disk/. I tried using other
# options such as requests-cache, but it was too slow compared to the solution
# implemented here.
# TODO(sarvjeets): It would be good to get rid of this one-off solution and switch
# to something more standard.


class Cacheable(ABC):
    @abstractmethod
    def CacheKey(self):
        """Unique string value used as key for caching."""
        pass


def get_file_age(file):
    return (datetime.today() -
            datetime.fromtimestamp(file.stat().st_mtime)).days


# Dict to keep cache context.
# cache_dir:
# The pathlib.Path object specifying cache directory. If set to None,
# caching is disabled. Default: ~/.lakshmicache
# force_refresh:
# If set to True, new values are generated even if a cached one is
# available.
_ctx = {'force_refresh': False}
_DEFAULT_DIR = Path.home() / '.lakshmicache'
_CACHE_STR = 'cache_dir'
_FORCE_STR = 'force_refresh'


def set_force_refresh(v):
    global _ctx
    _ctx[_FORCE_STR] = v


def set_cache_dir(cache_dir):
    """Sets the cache directory.

    If the cache directory is not specific, default ~/.lakshmicache
    is used.

    Args:
        cache_dir: The pathlib.Path object specifying cache directory.
        If set to None, caching is disabled.
    """
    global _ctx
    _ctx[_CACHE_STR] = cache_dir
    if cache_dir is None:
        return
    cache_dir.mkdir(exist_ok=True)
    for file in cache_dir.glob('*_*.lkc'):
        days = int(file.name.split('_')[0])
        if get_file_age(file) >= days:
            file.unlink()


def cache(days):
    def decorator(func):
        @functools.wraps(func)
        def new_func(class_obj):
            global _ctx
            if _CACHE_STR not in _ctx.keys():
                # Cache dir not set. Set to default.
                set_cache_dir(_DEFAULT_DIR)
            cache_dir = _ctx[_CACHE_STR]
            if not cache_dir:
                return func(class_obj)
            force_refresh = _ctx[_FORCE_STR]

            key = f'{func.__qualname__}_{class_obj.CacheKey()}'
            filename = f'{days}_{md5(key.encode("utf8")).hexdigest()}.lkc'
            file = cache_dir / filename

            if not force_refresh and file.exists() and get_file_age(file) < days:
                return pickle.loads(file.read_bytes())
            value = func(class_obj)
            file.write_bytes(pickle.dumps(value))
            return value
        return new_func
    return decorator



"""Module to help output and print tables."""

from tabulate import tabulate
import lakshmi.utils as utils

class Table():
    coltype2func = {
        'str': lambda x: x,
        'dollars': lambda x: utils.FormatMoney(x),
        'delta_dollars': lambda x: utils.FormatMoneyDelta(x),
        'percentage': lambda x: f'{round(100*x)}%',
        'float': lambda x: str(float(x)),
    }

    coltype2align = {
        'str': 'left',
        'dollars': 'right',
        'delta_dollars': 'right',
        'percentage': 'right',
        'float': 'decimal',
    }

    def __init__(self, numcols, headers=(), coltypes=None):
        assert numcols >= 0
        self._numcols = numcols

        if headers:
            assert len(headers) == numcols
        self._headers = headers

        if coltypes:
            assert len(coltypes) == numcols
            assert set(coltypes).issubset(
                self.coltype2func.keys()), 'Bad column type in coltypes'
            self._coltypes = coltypes
        else:
            self._coltypes = ['str'] * self._numcols

        self._rows = []

    def AddRow(self, row):
        assert len(row) <= self._numcols
        self._rows.append(row)
        return self

    def SetRows(self, rows):
        assert max(map(len, rows)) <= self._numcols
        self._rows = rows

    def Headers(self):
        return self._headers

    def ColAlign(self):
        return list(map(lambda x: self.coltype2align[x], self._coltypes))

    def List(self):
        return self._rows

    def StrList(self):
        ret_list = []
        for row in self.List():
            ret_row = []
            for col_num in range(len(row)):
                if row[col_num] is None:
                    ret_row.append('')
                else:
                    ret_row.append(self.coltype2func[self._coltypes[col_num]](
                        row[col_num]))
            ret_list.append(ret_row)
        return ret_list

    def String(self, tablefmt='simple'):
        str_list = self.StrList()
        if not str_list:
            return ''

        return tabulate(str_list,
                        headers=self.Headers(),
                        tablefmt=tablefmt,
                        colalign=self.ColAlign())



"""Implementation of standard asset types."""

from abc import ABC, abstractmethod
from lakshmi.cache import cache, Cacheable
import lakshmi.constants
from lakshmi.table import Table
import lakshmi.utils as utils
import datetime
import re
import requests
import yfinance


def ToDict(asset):
    return {asset.__class__.__name__: asset.ToDict()}

def FromDict(d):
    keys = list(d.keys())
    assert len(keys) == 1
    class_name = keys[0]

    for c in CLASSES:
        if c.__name__ == class_name:
            return c.FromDict(d.pop(class_name))

    raise AssertionError(f'Class {class_name} not found.')


class Asset(ABC):
    """Class representing an asset (fund, ETF, cash, etc.)."""
    def __init__(self, class2ratio):
        """
        Argments:
          class2ratio: Dict of class_name -> ratio. 0 < Ratio <= 1.0
        """
        self._delta = 0
        self.class2ratio = class2ratio

        total = 0
        for ratio in class2ratio.values():
            assert ratio >= 0.0 and ratio <= 1.0, (
                f'Bad Class ratio provided to Asset ({ratio})')
            total += ratio

        assert abs(total - 1.0) < 1e-6, (
            f'Total allocation to classes must be 100% (actual = {total * 100}%)')

    def ToDict(self):
        """Encodes this class into a dictionary.

        This method for non-abstract Asset classes encodes all data.
        This method for abstract Asset classes only encodes non-constructor data.
        """
        if self._delta != 0:
            return {'What if': self._delta}
        return dict()

    def FromDict(self, d):
        """Reverse of ToDict.

        This method for non-abstract Asset classes is a factory method.
        This method for abstract Asset classes decodes non-constructor data (if any).
        """
        self.WhatIf(d.pop('What if', 0))
        return self

    def ToTable(self):
        table = Table(2).AddRow(['Name:', f'{self.Name()}'])

        asset_mapping_table = Table(2, coltypes=['str', 'percentage'])
        for asset_class, ratio in self.class2ratio.items():
            asset_mapping_table.AddRow([f'{asset_class}', ratio])
        table.AddRow(['Asset Class Mapping:', f'{asset_mapping_table.String(tablefmt="plain")}'])

        if self._delta:
            table.AddRow(['Adjusted Value:', f'{utils.FormatMoney(self.AdjustedValue())}'])
            table.AddRow(['What if:', f'{utils.FormatMoneyDelta(self._delta)}'])
        else:
            table.AddRow(['Value:', f'{utils.FormatMoney(self.AdjustedValue())}'])

        return table

    def String(self):
        return self.ToTable().String(tablefmt='plain')

    def WhatIf(self, delta):
        if delta < 0 and delta < -self.AdjustedValue():
            delta = - self.AdjustedValue()
        self._delta += delta
        return delta

    def AdjustedValue(self):
        return self.Value() + self._delta

    @abstractmethod
    def Value(self):
        pass

    @abstractmethod
    def Name(self):
        pass

    @abstractmethod
    def ShortName(self):
        pass


class ManualAsset(Asset):
    def __init__(self, name, value, class2ratio):
        assert value >= 0, 'Value of an asset can not be negative.'
        self.name = name
        self.value = value
        super().__init__(class2ratio)

    def ToDict(self):
        d = {'Name': self.name,
             'Value': self.value,
             'Asset Mapping': self.class2ratio}
        d.update(super().ToDict())
        return d

    @classmethod
    def FromDict(cls, d):
        ret_obj = ManualAsset(d.pop('Name'),
                              d.pop('Value', 0),
                              d.pop('Asset Mapping'))
        Asset.FromDict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def Value(self):
        return self.value

    def Name(self):
        return self.name

    def ShortName(self):
        return self.name


class TaxLot:
    """Class to represent a single tax lot for an Asset."""

    def __init__(self, date, quantity, unit_cost):
        # Do some sanity check.
        date_pattern = re.compile('\\d{4}/\\d{2}/\\d{2}')
        assert date_pattern.match(
            date), 'Tax lot dates should be in format YYYY/MM/DD'

        self.date = date
        self.quantity = quantity
        self.unit_cost = unit_cost

    def ToDict(self):
        return {'Date': self.date,
                'Quantity': self.quantity,
                'Unit Cost': self.unit_cost}

    @classmethod
    def FromDict(cls, d):
        ret_obj = TaxLot(d.pop('Date'), d.pop('Quantity'), d.pop('Unit Cost'))
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj


class TradedAsset(Asset):
    """An abstract class representing an asset with units and per unit price."""

    def __init__(self, shares, class2ratio):
        self.shares = shares
        self.tax_lots = None
        super().__init__(class2ratio)

    def ToDict(self):
        d = dict()
        if self.tax_lots:
            d.update({'Tax Lots': [lot.ToDict() for lot in self.tax_lots]})
        d.update(super().ToDict())
        return d

    def FromDict(self, d):
        super().FromDict(d)
        if 'Tax Lots' not in d:
            return
        tax_lots_list = [TaxLot.FromDict(lot_dict)
                         for lot_dict in d.pop('Tax Lots')]
        self.SetLots(tax_lots_list)
        return self

    def SetLots(self, tax_lots_list):
        sum_lots = sum([t.quantity for t in tax_lots_list])
        assert abs(sum_lots - self.shares) < 1e-6, (
            f'Lots provided should sum up to {self.shares}')
        self.tax_lots = tax_lots_list
        return self

    def ListLots(self):
        table = Table(5,
                headers=['Date', 'Quanity', 'Cost', 'Gain', 'Gain%'],
                coltypes=['str', 'float', 'dollars', 'delta_dollars',
                    'percentage'])
        for lot in self.tax_lots:
            table.AddRow(
                    [lot.date,
                     lot.quantity,
                     lot.unit_cost * lot.quantity,
                     (self.Price() - lot.unit_cost) * lot.quantity,
                     self.Price() / lot.unit_cost - 1])
        return table

    def ToTable(self):
        table = super().ToTable()
        table.AddRow(['Price:', f'{utils.FormatMoney(self.Price())}'])
        return table

    def String(self):
        if not self.tax_lots:
            return super().String()

        return (super().String() + '\n\nTax lots:\n' +
                f'{self.ListLots().String()}')

    def Value(self):
        return self.shares * self.Price()

    # This class inherits abstract methods Name & ShortName from Asset.

    @abstractmethod
    def Price(self):
        pass


class NotFoundError(Exception):
    pass


class TickerAsset(TradedAsset, Cacheable):
    """An asset class representing a Ticker whose price can be pulled."""

    def __init__(self, ticker, shares, class2ratio):
        self.ticker = ticker
        session = requests.Session()
        session.headers['User-agent'] = (f'{lakshmi.constants.NAME}/'
            '{lakshmi.constants.VERSION}')
        self.yticker = yfinance.Ticker(ticker, session=session)
        super().__init__(shares, class2ratio)

    def ToDict(self):
        d = {'Ticker': self.ticker,
             'Shares': self.shares,
             'Asset Mapping': self.class2ratio}
        d.update(super().ToDict())
        return d

    @classmethod
    def FromDict(cls, d):
        ret_obj = TickerAsset(
            d.pop('Ticker'),
            d.pop('Shares'),
            d.pop('Asset Mapping'))
        TradedAsset.FromDict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def ToTable(self):
        table = super().ToTable()
        rows = table.List()
        rows.insert(0, ['Ticker:', f'{self.ticker}'])
        table.SetRows(rows)
        return table

    def CacheKey(self):
        return self.ticker

    @cache(365)  # Name changes are rare.
    def Name(self):
        if self.yticker.info.get('longName') is None:
            raise NotFoundError(
                f'Cannot retrieve ticker ("{self.ticker}") from Yahoo Finance')
        return self.yticker.info['longName']

    def ShortName(self):
        return self.ticker

    @cache(1)
    def Price(self):
        if self.yticker.info.get('regularMarketPrice') is None:
            raise NotFoundError(
                f'Cannot retrieve ticker ("{self.ticker}") from Yahoo Finance')
        return self.yticker.info['regularMarketPrice']


class VanguardFund(TradedAsset, Cacheable):
    """An asset class representing Vanguard trust fund represented by a numeric ID."""

    def __init__(self, fund_id, shares, class2ratio):
        self.fund_id = fund_id
        super().__init__(shares, class2ratio)

    def ToDict(self):
        d = {'Fund Id': self.fund_id,
             'Shares': self.shares,
             'Asset Mapping': self.class2ratio}
        d.update(super().ToDict())
        return d

    @classmethod
    def FromDict(cls, d):
        ret_obj = VanguardFund(
            d.pop('Fund Id'),
            d.pop('Shares'),
            d.pop('Asset Mapping'))
        TradedAsset.FromDict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def ToTable(self):
        table = super().ToTable()
        rows = table.List()
        rows.insert(0, ['Fund id:', f'{self.fund_id}'])
        table.SetRows(rows)
        return table

    def CacheKey(self):
        return str(self.fund_id)

    @cache(365)  # Name changes are very rare.
    def Name(self):
        req = requests.get(
            f'https://api.vanguard.com/rs/ire/01/pe/fund/{self.fund_id}/profile.json',
            headers={'Referer': 'https://vanguard.com/'})
        req.raise_for_status()  # Raise if error
        return req.json()['fundProfile']['longName']

    def ShortName(self):
        return str(self.fund_id)

    @cache(1)
    def Price(self):
        req = requests.get(
            f'https://api.vanguard.com/rs/ire/01/pe/fund/{self.fund_id}/price.json',
            headers={'Referer': 'https://vanguard.com/'})
        req.raise_for_status()  # Raise if error
        return float(req.json()['currentPrice']
                     ['dailyPrice']['regular']['price'])


class _TreasuryBonds(Asset):
    class Bond(Cacheable):
        """A class representing individual I or EE Bond."""
        def __init__(self, series, issue_date, denom):
            self.series = series
            self.issue_date = issue_date
            self.denom = denom
            self.redemption_date = datetime.datetime.now().strftime('%m/%Y')

        def CacheKey(self):
            return '{}_{}_{}'.format(
                self.series,
                self.issue_date.replace('/', '.'),
                self.redemption_date.replace('/', '.'))

        @cache(32)  # The value of a Bond doesn't change in a month.
        def _GetBondInfo(self):
            """Returns the rate and value of a $1000 bond."""
            data = {
                'RedemptionDate': self.redemption_date,
                'Series': self.series,
                'Denomination': '1000',
                'IssueDate': self.issue_date,
                'btnAdd.x': 'CALCULATE'
            }

            req = requests.post(
                'http://www.treasurydirect.gov/BC/SBCPrice', data=data)
            req.raise_for_status()

            ret_vals = re.findall('\n<td>.*</td>', req.text)
            rate = re.sub('\n|<[^>]+>', '', ret_vals[6])
            value = float(re.sub('\n|\\$|,|<[^>]+>', '', ret_vals[7]))
            # EE Bonds returned are half the value (I guess TD website
            # assumes paper bonds)
            return rate, value * (2.0 if self.series == 'EE' else 1.0)

        def Value(self):
            unused_rate, value = self._GetBondInfo()
            return value * (self.denom / 1000.0)

        def Rate(self):
            rate, unused_value = self._GetBondInfo()
            return rate

        def AsList(self):
            rate, value = self._GetBondInfo()
            value *= (self.denom / 1000.0)
            return [self.issue_date, self.denom, rate, value]

    def __init__(self, series, class2ratio):
        self.series = series
        super().__init__(class2ratio)
        self.bonds = []

    def ToDict(self):
        d = {}
        d['Bonds'] = []
        for bond in self.bonds:
            d['Bonds'].append(
                {'Issue Date': bond.issue_date, 'Denomination': bond.denom})
        d.update(super().ToDict())
        return d

    def FromDict(self, d):
        for bond in d.pop('Bonds'):
            self.AddBond(bond.pop('Issue Date'), bond.pop('Denomination'))
            assert len(bond) == 0, f'Extra attributes found: {list(bond.keys())}'
        Asset.FromDict(self, d)
        return self

    def AddBond(self, issue_date, denom):
        self.bonds.append(self.Bond(self.series, issue_date, denom))
        return self

    def Value(self):
        value = 0.0
        for bond in self.bonds:
            value += bond.Value()
        return value

    def ListBonds(self):
        table = Table(
                4,
                headers=['Issue Date', 'Denom', 'Rate', 'Value'],
                coltypes=['str', 'dollars', 'str', 'dollars'])
        for bond in self.bonds:
            table.AddRow(bond.AsList())
        return table

    def String(self):
        return (super().String() + '\n\nBonds:\n' +
                f'{self.ListBonds().String()}')


class IBonds(_TreasuryBonds):
    def __init__(self, class2ratio):
        super().__init__('I', class2ratio)

    def ToDict(self):
        d = {'Asset Mapping': self.class2ratio}
        d.update(super().ToDict())
        return d

    @classmethod
    def FromDict(cls, d):
        ret_obj = IBonds(d.pop('Asset Mapping'))
        _TreasuryBonds.FromDict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def Name(self):
        return 'I Bonds'

    def ShortName(self):
        return self.Name()


class EEBonds(_TreasuryBonds):
    def __init__(self, class2ratio):
        super().__init__('EE', class2ratio)

    def ToDict(self):
        d = {'Asset Mapping': self.class2ratio}
        d.update(super().ToDict())
        return d

    @classmethod
    def FromDict(cls, d):
        ret_obj = EEBonds(d.pop('Asset Mapping'))
        _TreasuryBonds.FromDict(ret_obj, d)
        assert len(d) == 0, f'Extra attributes found: {list(d.keys())}'
        return ret_obj

    def Name(self):
        return 'EE Bonds'

    def ShortName(self):
        return self.Name()


CLASSES = [ManualAsset, TickerAsset, VanguardFund, IBonds, EEBonds]



"""Classes to analyze the portfolio.

These classes are not meant to change the portfolio. They are used for
just analyzing the portfolio and output the results (if any).
"""

from abc import ABC, abstractmethod
from lakshmi.table import Table


class Analyzer(ABC):
    @abstractmethod
    def Analyze(self, portfolio):
        pass


class TLH(Analyzer):
    """Tax loss harvesting opportunity analyzer.

    This triggers a message with details of tax lots that could be sold to
    harvest the losses.
    """

    def __init__(self, max_percentage, max_dollars=None):
        """Triggers if a lot has lost more than max_percentage or the total loss
        for a given asset exceeds max_dollars.
        """
        assert max_percentage > 0.0 and max_percentage < 1.0, (
            'max_percetage should be between 0% and 100% (exclusive).')
        self.max_percentage = max_percentage
        if max_dollars:
            assert max_dollars > 0, 'max_dollars should be positive.'
        self.max_dollars = max_dollars

    def _ReturnLotsToSell(self, price, tax_lots):
        percent_lots = []
        negative_lots = []
        total_loss = 0.0

        for lot in tax_lots:
            loss = (lot.unit_cost - price) * lot.quantity
            loss_percent = (lot.unit_cost - price) / lot.unit_cost
            if loss > 0:
                negative_lots.append([lot.date, loss, loss_percent])
                total_loss += loss
            if loss_percent > self.max_percentage:
                percent_lots.append([lot.date, loss, loss_percent])

        if self.max_dollars is not None and total_loss > self.max_dollars:
            return negative_lots
        else:
            return percent_lots

    def Analyze(self, portfolio):
        ret_val = Table(
            5,
            headers=['Account', 'Asset', 'Date', 'Loss', 'Loss%'],
            coltypes=['str', 'str', 'str', 'dollars', 'percentage'])
        for account in portfolio.Accounts():
            for asset in account.Assets():
                if hasattr(asset, 'tax_lots') and asset.tax_lots:
                    assert hasattr(
                        asset, 'Price'), 'Asset has tax lots but no Price'
                    price = asset.Price()
                    for lot in self._ReturnLotsToSell(price, asset.tax_lots):
                        ret_val.AddRow(
                            [account.Name(), asset.ShortName()] + lot)
        return ret_val


class BandRebalance(Analyzer):
    """Triggers if portfolio asset class targets are outside the bands
    specified. This considers an asset class outside bound if the
    absolute difference in percentage allocation is more than max_abs_percent
    different from target allocation or more than max_relative_percent
    different from the target allocation."""

    def __init__(self, max_abs_percent=0.05, max_relative_percent=0.25):
        assert max_abs_percent > 0 and max_abs_percent < 1.0
        assert max_relative_percent > 0 and max_relative_percent < 1.0
        self.max_abs_percent = max_abs_percent
        self.max_relative_percent = max_relative_percent

    def Analyze(self, portfolio):
        aa = portfolio.AssetAllocation(portfolio.asset_classes.Leaves())
        headers = ['Class', 'Actual%', 'Desired%', 'Value', 'Difference']
        assert headers == aa.Headers()
        ret_val = Table(
            5,
            headers,
            ['str', 'percentage', 'percentage', 'dollars', 'delta_dollars'])
        for row in aa.List():
            abs_percent = abs(row[1] - row[2])
            rel_percent = abs_percent / row[2] if row[2] != 0 else 0
            if (abs_percent >= self.max_abs_percent or
                    rel_percent >= self.max_relative_percent):
                ret_val.AddRow(row)

        return ret_val



"""Command line interface for Lakshmi.

This is meant to be used as an application and would not work well if used
as library (by design it keeps a lot of global state and is not safe to be
called multiple times from the same program). If there is ever need to use
it as a library, this code requires major refactoring to clean it up."""

from pathlib import Path
from lakshmi import Portfolio
from lakshmi.table import Table
import click
import lakshmi.analyze
import lakshmi.assets
import lakshmi.cache
import yaml


class LakContext:
    """Context class with utilities to help the script keep state and
    share it."""
    DEFAULT_PORTFOLIO = '~/portfolio.yaml'

    def _ReturnConfig(self):
        """Internal function to read and return .lakrc file."""
        lakrcfile = Path(self.lakrc).expanduser()

        if not lakrcfile.exists():
            return {}
        config = yaml.load(lakrcfile.read_text(), Loader=yaml.SafeLoader)
        return config

    def __init__(self, lakrc='~/.lakrc'):
        self.lakrc = lakrc
        # Used in self.OptionalSeparator()
        self.continued = False
        # List of hypothetical whatifs. Used in self.GetWhatIfs() and
        # self.WarnForWhatIfs()
        self.whatifs = None
        # The loaded portfolio.
        self.portfolio = None
        self.tablefmt = None

        config = self._ReturnConfig()

        portfolio_filename = config.pop(
                'portfolio', LakContext.DEFAULT_PORTFOLIO)
        self.portfolio_filename = str(Path(portfolio_filename).expanduser())

        # Setup cache directory.
        if 'cache' in config:
            cache_dir = config.pop('cache')
            if cache_dir is None:
                # Disable caching.
                lakshmi.cache.set_cache_dir(None)
            else:
                lakshmi.cache.set_cache_dir(Path(cache_dir).expanduser())

        if len(config):
            raise click.ClickException(f'Extra entries found in config file: '
                                       '{list(config.keys())}')

    def OptionalSeparator(self):
        """Prints a newline between multiple commands. Used to add a newline
        between cmd1, cmd2, cmd3, etc. in 'lak list cmd1 cmd2 cmd3'."""
        # Don't print separator if this is the first time we are called.
        if self.continued:
            click.echo()
        # Set it up so that separator is printed for the next command.
        self.continued = True

    def GetWhatIfs(self):
        """Load and return a list of hypothetical whatifs in the portfolio."""
        if not self.whatifs:
            self.whatifs = self.Portfolio().GetWhatIfs()
        return self.whatifs

    def WarnForWhatIfs(self):
        """Prints a warning if whatifs are set."""
        # Make sure we don't print warning multiple times if commands are
        # chained.
        if self.continued:
            return

        self.GetWhatIfs()
        if self.whatifs[0].List() or self.whatifs[1].List():
            click.secho('Warning: Hypothetical what ifs are set.\n', fg='red')

    def Portfolio(self):
        """Loads and returns the portfolio from self.portfolio_filename."""
        if not self.portfolio:
            # Check if portfolio file doesn't exist and print helpful error
            # message.
            portfolio_file = Path(self.portfolio_filename).expanduser()
            if not portfolio_file.exists():
                raise click.ClickException(
                    f'Portfolio file {portfolio_file} does not exist. Please '
                    'use "lak init" to create a new portfolio.')

            self.portfolio = Portfolio.Load(self.portfolio_filename)
        return self.portfolio

    def SavePortfolio(self):
        """Save self.portfolio back to file."""
        self.portfolio.Save(self.portfolio_filename)


# Global variable to save and pass context between click commands.
#
# I tried using click's builtin context, but it was too troublesome
# and didn't exactly give us the functionality that I wanted
# -- sarvjeets
lakctx = None


@click.group()
@click.version_option()
@click.option('--refresh', '-r', is_flag=True,
              help='Re-fetch all data instead of using previously cached '
              'data. For large portfolios this would be extremely slow.')
def lak(refresh):
    lakshmi.cache.set_force_refresh(refresh)
    global lakctx
    if not lakctx:
        # Setup a new context object for child commands.
        lakctx = LakContext()

@lak.group(chain=True)
@click.option('--format',  '-f',
              type=click.Choice(
                  ['plain', 'simple', 'github', 'grid', 'fancy_grid',
                    'pipe', 'orgtbl', 'rst', 'mediawiki', 'html', 'latex',
                    'latex_raw', 'latex_booktabs', 'latex_longtable', 'tsv'],
                  case_sensitive=False),
              default='simple',
              help='Set output table format. For more information on table'
              'formats, please see "Table format" section on: '
              'https://pypi.org/project/tabulate/')
def list(format):
    """Command to list various parts of the portfolio."""
    global lakctx
    lakctx.tablefmt = format


@list.command()
def total():
    """Prints the total value of the portfolio."""
    global lakctx
    lakctx.WarnForWhatIfs()
    lakctx.OptionalSeparator()
    portfolio = lakctx.Portfolio()
    click.echo(
        Table(2, coltypes=['str', 'dollars']).AddRow(
            ['Total Assets', portfolio.TotalValue()]).String(lakctx.tablefmt))


@list.command()
def al():
    """Prints the Asset Location of the portfolio. For more information,
    please see
    https://www.bogleheads.org/wiki/Tax-efficient_fund_placement"""
    global lakctx
    lakctx.WarnForWhatIfs()
    lakctx.OptionalSeparator()
    portfolio = lakctx.Portfolio()
    click.echo(portfolio.AssetLocation().String(lakctx.tablefmt))


@list.command()
@click.option('--compact/--no-compact', default=True, show_default=True,
              help='Print the Asset allocation tree in a vertically '
              'compact format')
@click.option('--asset-class', '-a', type=str,
              help='If provided, only print asset allocation for these asset '
              'classes. This is comma seperated list of asset classes (not '
              'necessarily leaf asset classes) and the allocation across '
              'these asset classes should sum to 100%.')
def aa(compact, asset_class):
    """Prints the Asset Allocation of the portfolio. For more information,
    please see https://www.bogleheads.org/wiki/Asset_allocation"""
    global lakctx
    lakctx.WarnForWhatIfs()
    lakctx.OptionalSeparator()

    portfolio = lakctx.Portfolio()
    if asset_class:
        if not compact:
            raise click.UsageError(
                '--no-compact is only supported when --asset-class '
                'is not specified.')
        classes_list = [c.strip() for c in asset_class.split(',')]
        click.echo(portfolio.AssetAllocation(classes_list).String(
            lakctx.tablefmt))
    else:
        if compact:
            click.echo(portfolio.AssetAllocationCompact().String(
                lakctx.tablefmt))
        else:
            click.echo(portfolio.AssetAllocationTree().String(
                lakctx.tablefmt))


@list.command()
@click.option('--short-name', '-s', is_flag=True,
              help='Print the short name of the assets as well (e.g. Ticker '
              'for assets that have it).')
@click.option('--quantity', '-q', is_flag=True,
              help='Print the quantity of the asset (e.g. quantity/shares '
              'for assets that have it).')
def assets(short_name, quantity):
    """Prints all assets in the portfolio and their current values."""
    global lakctx
    lakctx.WarnForWhatIfs()
    lakctx.OptionalSeparator()
    portfolio = lakctx.Portfolio()
    click.echo(portfolio.Assets(short_name=short_name,
                                quantity=quantity).String(lakctx.tablefmt))


@list.command()
def whatifs():
    """Print hypothetical what ifs for assets and accounts."""
    global lakctx
    account_whatifs, asset_whatifs = lakctx.GetWhatIfs()
    if account_whatifs.List():
        lakctx.OptionalSeparator()
        click.echo(account_whatifs.String(lakctx.tablefmt))
    if asset_whatifs.List():
        lakctx.OptionalSeparator()
        click.echo(asset_whatifs.String(lakctx.tablefmt))


@lak.command(context_settings={"ignore_unknown_options": True})
@click.option('--asset', '-a', type=str, metavar='substr',
              help='Make changes to this asset (a sub-string that matches '
              'either the asset name or the short name)')
@click.option('--account', '-t', type=str, metavar='substr',
              help='Make changes to this account (a sub-string that matches '
              'the account name)')
@click.option('--reset', '-r', is_flag=True,
              help='Reset all hypothetical whatif amounts.')
@click.argument('delta', type=float, required=False)
def whatif(asset, account, reset, delta):
    """Run hypothetical what if scenarios by adding DELTA to an account
    or asset. This is useful to see how the asset allocation or location
    will change if you make some changes without *actually* making those
    changes. Once you are done playing around with the hypothetical changes,
    you can reset them all by using --reset flag."""
    global lakctx
    portfolio = lakctx.Portfolio()

    if reset:
        # Sanity check the flags.
        if asset or account or delta:
            raise click.UsageError(
                'Can\'t specify any other flags/arguments when --reset is '
                'specified')
        portfolio.ResetWhatIfs()
        lakctx.SavePortfolio()
        return

    if not delta:
        raise click.UsageError('Must specify a value (delta) to add/subtract '
                               'from an account or asset')

    if not asset and account:  # Delta is applied to Account
        account_name = portfolio.GetAccountNameBySubStr(account)
        portfolio.WhatIfAddCash(account_name, delta)
        lakctx.SavePortfolio()
        return

    if asset:
        account_name, asset_name = portfolio.GetAssetNameBySubStr(
            account if account is not None else '', asset)
        portfolio.WhatIf(account_name, asset_name, delta)
        lakctx.SavePortfolio()
        return

    raise click.UsageError('At least one of --asset or --account must '
                           'be specified')


@lak.command()
@click.option('--asset', '-a', type=str, metavar='substr',
              help='Get Info about this asset (a sub-string that matches '
              'either the asset name or the short name)')
@click.option('--account', '-t', type=str, metavar='substr',
              help='Get Info about this account (a sub-string that matches '
              'the account name)')
def info(asset, account):
    """Prints detailed information about an asset or account. If only account
    is specified, this command prints information about an account. If asset
    or both asset and acccount is specified, it prints information about an
    asset. In the second case, asset (and optionally account) must match
    exactly one asset/account pair in the portfolio."""
    global lakctx
    portfolio = lakctx.Portfolio()

    if not asset and account:  # Print info for account.
        account_name = portfolio.GetAccountNameBySubStr(account)
        click.echo(portfolio.GetAccount(account_name).String())
        return

    if asset:  # Print info for asset.
        account_name, asset_name = portfolio.GetAssetNameBySubStr(
            account if account is not None else '', asset)
        click.echo(portfolio.GetAccount(
            account_name).GetAsset(asset_name).String())
        return

    raise click.UsageError('At least one of --asset or --account must '
                           'be specified')


@lak.group()
def edit():
    """Edit parts of the portfolio."""
    pass


def edit_and_parse(edit_dict, parse_fn, filename):
    """Helper funtion to edit parts of portfolio.

    This function is used to add/edit parts of the portfolio. It converts
    a dictionary representation of the object into a YAML file that the
    user can edit. The file is then converted back to a dictionary, which
    is parsed via the parse_fn to convert it into an object.

    Arguments:
        edit_dict: Dictionary representing an object that is being
        editted. This is usually the output of ToDict() of an object.
        parse_fn: The function used to parse the resulting dict generated
        by parsing the yaml file.
        filename: The file containing a template YAML file for the object
        being editted (inside lakshmi/data directory). If edit_dict is
        empty, this file is presented to the user so that the user can make
        changes inline. If edit_dict is set, this file is added as a
        comment at the end of the edit_dict representation as a helping
        guide for the user.
    """
    # Change filename to absolute path.
    filepath = (Path(__file__).parents[2].absolute() /
                'data' / filename)
    if edit_dict:
        HELP_MSG = ('\n\n' + '# # Lines starting with "#" are ignored and an '
                    'empty message aborts this command.\n\n' +
                    filepath.read_text().replace('\n', '\n# '))
        edit_str = yaml.dump(edit_dict, sort_keys=False) + HELP_MSG
    else:
        edit_str = filepath.read_text()

    while True:
        edit_str = click.edit(edit_str)
        if not edit_str:  # No changes or empty string.
            raise click.Abort()
        try:
            parse_str = (edit_str.split(HELP_MSG, 1)[0].rstrip('\n')
                         if edit_dict else edit_str)
            return parse_fn(yaml.load(parse_str, Loader=yaml.SafeLoader))
        except Exception as e:
            click.echo('Error parsing file: ' + repr(e))
            if not click.confirm('Do you want to edit again?'):
                raise click.Abort()


@lak.command()
def init():
    """Initializes a new portfolio by adding asset classes. This command can
    be used to create an empty portfolio file if one doesn't exist."""
    global lakctx
    if Path(lakctx.portfolio_filename).exists():
        raise click.ClickException('Portfolio file already exists: ' +
                                   lakctx.portfolio_filename)

    asset_class = edit_and_parse(
        None,
        lambda x: lakshmi.AssetClass.FromDict(x).Validate(),
        'AssetClass.yaml')
    Portfolio(asset_class).Save(lakctx.portfolio_filename)


@edit.command()
def assetclass():
    """Edit the Asset classes and desired asset allocation across
    those asset classes."""
    global lakctx
    portfolio = lakctx.Portfolio()

    asset_classes = edit_and_parse(
        portfolio.asset_classes.ToDict(),
        lambda x: lakshmi.AssetClass.FromDict(x).Validate(),
        'AssetClass.yaml')
    portfolio.asset_classes = asset_classes
    lakctx.SavePortfolio()


@edit.command()
@click.option('--account', '-t', type=str, metavar='substr', required=True,
              help='Edit the account that matches this sub-string')
def account(account):
    """Edit an account in the portfolio."""
    global lakctx
    portfolio = lakctx.Portfolio()

    account_name = portfolio.GetAccountNameBySubStr(account)
    account_obj = portfolio.GetAccount(account_name)
    assets = account_obj.Assets()
    account_obj.SetAssets([])

    account_obj = edit_and_parse(
        account_obj.ToDict(),
        lakshmi.Account.FromDict,
        'Account.yaml')
    account_obj.SetAssets(assets)

    if account_obj.Name() != account_name:
        portfolio.RemoveAccount(account_name)
    portfolio.AddAccount(account_obj, replace=True)
    lakctx.SavePortfolio()


@edit.command()
@click.option('--asset', '-a', type=str, metavar='substr', required=True,
              help='Get Info about this asset (a sub-string that matches '
              'either the asset name or the short name).')
@click.option('--account', '-t', type=str, metavar='substr',
              help='Get Info about this account (a sub-string that matches the '
              'account name).')
def asset(asset, account):
    """Edit an asset in the portfolio."""
    global lakctx
    portfolio = lakctx.Portfolio()

    account_name, asset_name = portfolio.GetAssetNameBySubStr(
        account if account is not None else '', asset)
    account_obj = portfolio.GetAccount(account_name)
    asset_obj = account_obj.GetAsset(asset_name)

    asset_obj = edit_and_parse(asset_obj.ToDict(),
                               asset_obj.FromDict,
                               asset_obj.__class__.__name__ + '.yaml')

    if asset_obj.ShortName() != asset_name:
        account_obj.RemoveAsset(asset_name)
    account_obj.AddAsset(asset_obj, replace=True)

    lakctx.SavePortfolio()


@lak.group()
def add():
    """Add new accounts or assets to the portfolio."""
    pass


@add.command()
def account():
    """Add a new account to the portfolio."""
    global lakctx
    portfolio = lakctx.Portfolio()

    account = edit_and_parse(None, lakshmi.Account.FromDict, 'Account.yaml')
    portfolio.AddAccount(account)
    lakctx.SavePortfolio()


@add.command()
@click.option('--asset-type', '-p', required=True,
              type=click.Choice([c.__name__ for c in lakshmi.assets.CLASSES],
                                case_sensitive=False),
              help='Add this type of asset.')
@click.option('--account', '-t', type=str, metavar='substr', required=True,
              help='Add asset to this account (a sub-string that matches the '
              'account name).')
def asset(asset_type, account):
    """Edit assets in the portfolio."""
    global lakctx
    portfolio = lakctx.Portfolio()

    account_name = portfolio.GetAccountNameBySubStr(account)
    account_obj = portfolio.GetAccount(account_name)
    asset_cls = [
        c for c in lakshmi.assets.CLASSES if c.__name__ == asset_type][0]

    asset_obj = edit_and_parse(None, asset_cls.FromDict, asset_type + '.yaml')
    account_obj.AddAsset(asset_obj)

    lakctx.SavePortfolio()


@lak.group()
def delete():
    """Delete an account or asset."""
    pass


@delete.command()
@click.option('--account', '-t', type=str, metavar='substr', required=True,
              help='Delete the account that matches this sub-string')
@click.confirmation_option(prompt='This operation is not reversable. '
                           'Are you sure?')
def account(account):
    """Delete an account in the portfolio."""
    global lakctx
    portfolio = lakctx.Portfolio()
    account_name = portfolio.GetAccountNameBySubStr(account)
    portfolio.RemoveAccount(account_name)
    lakctx.SavePortfolio()


@delete.command()
@click.option('--asset', '-a', type=str, metavar='substr', required=True,
              help='Delete this asset (a sub-string that matches either '
              'the asset name or the short name).')
@click.option('--account', '-t', type=str, metavar='substr',
              help='Delete the specified asset from this account (a sub-string '
              'that matches the account name).')
@click.confirmation_option(prompt='This operation is not reversable. '
                           'Are you sure?')
def asset(asset, account):
    """Delete an asset from the portfolio."""
    global lakctx
    portfolio = lakctx.Portfolio()

    account_name, asset_name = portfolio.GetAssetNameBySubStr(
        account if account is not None else '', asset)
    account_obj = portfolio.GetAccount(account_name)
    account_obj.RemoveAsset(asset_name)
    lakctx.SavePortfolio()


@lak.group(chain=True)
def analyze():
    """Analyze the portfolio. """
    pass


@analyze.command()
@click.option('--max-percentage', '-p', type=float, default=10,
              show_default=True,
              help='The max percenatage loss for each lot before TLHing.')
@click.option('--max-dollars', '-d', type=int,
              help='The max absolute loss for an asset (across all tax lots) '
              'before TLHing.')
def tlh(max_percentage, max_dollars):
    """Shows which tax lots can be Tax-loss harvested."""
    global lakctx
    lakctx.OptionalSeparator()
    table = lakshmi.analyze.TLH(max_percentage / 100, max_dollars).Analyze(
        lakctx.Portfolio())
    if not table.List():
        click.echo('No tax lots to harvest.')
    else:
        click.echo(table.String())


@analyze.command()
@click.option('--max-abs-percentage', '-a', type=float, default=5,
              show_default=True,
              help='Max absolute difference before rebalancing.')
@click.option('--max-relative-percentage', '-r', type=float, default=25,
              help='The max relatve differnce before rebalancing.')
def rebalance(max_abs_percentage, max_relative_percentage):
    """Shows if assets needs to be rebalanced based on a band
    based rebalancing scheme. For more information, please refer to
    https://www.whitecoatinvestor.com/rebalancing-the-525-rule/."""
    global lakctx
    lakctx.OptionalSeparator()
    table = lakshmi.analyze.BandRebalance(
        max_abs_percentage / 100, max_relative_percentage / 100).Analyze(
        lakctx.Portfolio())
    if not table.List():
        click.echo('Portfolio Asset allocation within bounds.')
    else:
        click.echo(table.String())


if __name__ == '__main__':
    lak()



from setuptools import setup, find_packages
from lakshmi.constants import NAME, VERSION

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name=NAME,
    version=VERSION,
    author='Sarvjeet Singh',
    author_email='sarvjeet@gmail.com',
    description='Investing Tools.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sarvjeets/lakshmi',
    platforms='any',
    py_modules=['lakshmi'],
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    install_requires=[
        'click~=8.0.1',
        'PyYAML~=5.4.1',
        'requests~=2.25.1',
        'tabulate~=0.8.9',
        'yfinance~=0.1.59',
    ],
    entry_points={
        'console_scripts': [
            'lak = lakshmi.lak:lak',
        ],
    },
    test_suite='tests',
    python_requires='>=3.9.5',
)
